import traceback
from typing import Any

from attrs import define


@define
class ViewModel:
    _model: object
    _extra_attrs: dict[str, Any] | None = None
    _wrapped = True

    @classmethod
    def from_many(cls, objects: list) -> list:
        return [cls(obj) for obj in objects]

    def __getattr__(self, key):
        if self._extra_attrs is None:
            try:
                self._extra_attrs = self.extra_attrs()
            except Exception as e:
                traceback.print_exc()
                raise RuntimeError from e

        if key in self._extra_attrs:
            return self._extra_attrs[key]

        value = getattr(self._model, key)
        return value

    def extra_attrs(self):
        return {}

    def _unwrap(self):
        return self._model


def unwrap(obj):
    if hasattr(obj, "_model"):
        return obj._model
    return obj



"""Main module."""
import types
from collections import defaultdict

from flask import Flask, g

from pagic.page import Page, Route
from pagic.routing import url_for


class Pagic:
    app: Flask | None
    roots: list
    all_page_classes: list

    def __init__(self, app: Flask | None = None):
        self.roots = []
        self.all_page_classes = []
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask):
        if "pagic" in app.extensions:
            raise RuntimeError(
                "This extension is already registered on this Flask app."
            )

        self.app = app
        app.extensions["pagic"] = self

        app.before_request(self.before_request)
        app.context_processor(self.inject_extra_context)
        app.template_global("url_for")(url_for)

    def before_request(self):
        g.menus = defaultdict(list)

        for page_class in self.all_page_classes:
            menu_name = page_class.menu
            if not menu_name:
                continue

            page: Page = page_class()
            endpoint = page.endpoint
            label = page.label
            url = url_for(endpoint)
            menu_item = {
                "label": label,
                "endpoint": endpoint,
                "url": url,
                "active": False,
            }
            g.menus[menu_name].append(menu_item)

    def inject_extra_context(self):
        return {
            "url_for": url_for,
            "pagic": self,
        }

    # TODO
    # def register_macros(app, path):
    #     scanner = venusian.Scanner()
    #     scanner.scan(importlib.import_module(path))
    #
    #     for macro in MACROS:
    #         app.template_global(macro.__name__)(macro)

    #
    # New registration API
    #
    def register_roots(self, roots):
        self.roots = roots
        for page_class in roots:
            self.register_page(page_class)

    def register_page(self, page_class, ancestors: list | None = None):
        self.all_page_classes.append(page_class)

        if ancestors is None:
            ancestors = []

        methods = ["GET", "POST"]
        route = Route(page_class)

        # if hasattr(page_class, "routes"):
        #     for _route in page_class.routes:
        #         self.app.add_url_rule(_route, page_class.name, route, methods=methods)
        #     return

        if page_class.path is None:
            page_class.path = page_class.name

        path_list = []
        for p in ancestors + [page_class]:
            if p.path:
                path_list.append(p.path)
        path = "/" + "/".join(path_list)
        self.app.add_url_rule(path, route.endpoint, route, methods=methods)

        if hasattr(page_class, "children"):
            for child_class in page_class.children:
                self.register_page(child_class, ancestors + [page_class])

        # if hasattr(cls, "routes"):
        #     for _route in cls.routes:
        #         blueprint.add_url_rule(_route, route.endpoint, route, methods=methods)
        #     return
        #
        # if not hasattr(cls, "path"):
        #     cls.path = cls.name

        # blueprint.add_url_rule(cls.path, route.endpoint, route, methods=methods)

        for method_name in self.get_exposed_method_names(page_class):
            route = Route(page_class, method_name)
            self.app.add_url_rule(route.path, route.endpoint, route, methods=methods)

    @staticmethod
    def get_exposed_method_names(page_class):
        result = []
        for name, method in vars(page_class).items():
            if not isinstance(method, types.FunctionType):
                continue

            if not hasattr(method, "_pagic_metadata"):
                continue

            metadata = method._pagic_metadata  # type: ignore
            if not metadata.get("exposed"):
                continue

            result.append(name)

        return result

    #
    # Old API
    #
    # def register_pages(self, path=None):
    #     self.scan_pages(path)
    #
    #     for page_cls in Page.__all__pages__.values():
    #         print(f"Registering pags: {page_cls}")
    #         self.register_page(page_cls)
    #
    # def scan_pages(self, path):
    #     if not path:
    #         path = "tests"
    #     module = importlib.import_module(path)
    #     scanner = venusian.Scanner()
    #     scanner.scan(module)
    #
    # def register_page(self, cls):
    #     methods = ["GET", "POST"]
    #
    #     route = Route(cls)
    #     if hasattr(cls, "routes"):
    #         for _route in cls.routes:
    #             self.app.add_url_rule(_route, cls.name, route, methods=methods)
    #         return
    #
    #     if not hasattr(cls, "path"):
    #         cls.path = cls.name
    #
    #     self.app.add_url_rule(cls.path, cls.name, route, methods=methods)



from dataclasses import dataclass
from typing import Any, Sequence

from flask import render_template, request, url_for
from werkzeug.exceptions import MethodNotAllowed

# from app.flask.lib.view_model import unwrap
# from app.services.json_ld import to_json_ld
# from app.services.opengraph import to_opengraph
# from pagic import url_for

PAGES = {}


def fqdn(cls):
    return f"{cls.__module__}.{cls.__name__}"


def page(cls):
    Page.__all__pages__[fqdn(cls)] = cls
    return cls


def expose(method):
    if not hasattr(method, "_pagic_metadata"):
        method._pagic_metadata = {}
    method._pagic_metadata["exposed"] = True
    return method


class Page:
    __all__pages__ = {}

    name: str
    endpoint: str
    label: str
    path: str | None = None
    layout: str = ""
    menu: str = ""
    parent: Any = None
    children: Sequence = []
    args: dict = {}
    path_args: dict = {}
    query_args: dict = {}
    form_data: dict = {}

    @property
    def template(self) -> str:
        return f"pages/{self.name}.j2"

    @property
    def url(self):
        args = {}
        args.update(self.path_args)
        args.update(self.query_args)
        return url_for(self.endpoint, **args)

    @property
    def endpoint(self):
        return self.name

    @property
    def label(self):
        return self.name.capitalize()

    def context(self):
        """Override in subclasses."""
        return {}

    def get(self):
        return self.render()

    def post(self):
        """Override in subclasses, if needed."""
        raise MethodNotAllowed()

    def render(self):
        ctx = self.context()
        ctx.update(self.extra_context(ctx))
        content = self.content(ctx)
        if not self.layout:
            return content

        ctx["content"] = content
        return render_template(self.layout, **ctx)

    def content(self, ctx):
        return render_template(self.template, **ctx)

    def extra_context(self, ctx):
        d = {}

        if "title" not in ctx:
            d["title"] = self.label

        d["breadcrumbs"] = self.breadcrumbs

        d["json_data"] = {}

        # menus = {
        #     "main": g.nav_data["main"],
        #     "user": g.nav_data["user"],
        # }
        # menus.update(self.menus())
        # d["menus"] = menus
        #
        # if hasattr(self, "view_model") and "model" not in d:
        #     d["model"] = self.view_model

        # if "model" in d:
        #     model = unwrap(d["model"])
        #     d["og_data"] = to_opengraph(model)
        #     d["json_ld"] = to_json_ld(model)

        return d

    def menus(self):
        return {}

    @property
    def breadcrumbs(self):
        breadcrumbs = [
            {"name": self.label, "href": self.url, "current": True},
        ]
        if not self.parent:
            return breadcrumbs

        parent = self.parent()
        while True:
            breadcrumbs += [
                {"name": parent.label, "href": parent.url, "current": False},
            ]
            if not parent.parent:
                break
            parent = parent.parent()

        breadcrumbs.reverse()
        return breadcrumbs


@dataclass
class Route:
    __name__ = "Route"

    def __init__(self, page_class, method_name=""):
        self.page_class = page_class
        self.method_name = method_name

    def __call__(self, **kwargs):
        page = self.page_class()
        page.path_args = kwargs
        page.query_args = request.args
        page.form_data = request.form

        page.args = {}
        page.args.update(page.query_args)
        page.args.update(page.form_data)
        page.args.update(page.path_args)

        if self.method_name:
            return getattr(page, self.method_name)()

        match request.method:
            case "GET":
                return page.get()
            case "POST":
                return page.post()

    @property
    def path(self):
        page_class = self.page_class
        if page_class.path.startswith("/"):
            path = page_class.path
        else:
            path = "/" + page_class.path

        if self.method_name:
            return path + "/" + self.method_name
        else:
            return path

    @property
    def endpoint(self):
        if self.method_name:
            return self.page_class.name + "__" + self.method_name
        else:
            return self.page_class.name

    # @property
    # def __name__(self):
    #     if self.method_name:
    #         return self.page_class.__name__ + "__" + self.method_name
    #     else:
    #         return self.page_class.__name__



import pytest
from flask import Flask

from pagic.page import Page, page
from pagic.pagic import Pagic


@page
class HomePage(Page):
    name = "home"
    path = "/"
    label = "Home"

    layout = "tests/base.j2"
    template = "tests/home.j2"


@pytest.fixture
def app():
    app = Flask(__name__)
    pagic = Pagic(app)
    pagic.register_roots([HomePage])
    return app


def test_home(app, client):
    # rules = list(app.url_map.iter_rules())
    res = client.get("/")
    assert res.status_code == 200


if __name__ == "__main__":
    _app = Flask(__name__)
    pagic = Pagic(_app)
    _app.run()



from functools import singledispatch

from flask import url_for as url_for_orig
from werkzeug.routing import BuildError

from pagic import Page


@singledispatch
def url_for(obj, _ns="", **kw) -> str:
    # print(f"Illegal argument for 'url_for': {obj} (type: {type(obj)})")
    # return "#"

    # Was:
    raise BuildError(f"Illegal argument for 'url_for': {obj} (type: {type(obj)})")


@url_for.register
def url_for_page(page: Page, _ns="", **kw):
    return page.url


@url_for.register
def url_for_str(name: str, _ns="", **kw):
    return url_for_orig(name, **kw)



from flask import request

from pagic.routing import url_for


def make_menu(menu_spec):
    path = request.path
    menu = []
    for d in menu_spec:
        endpoint = d["endpoint"]
        if endpoint.startswith("/"):
            url = endpoint
        else:
            url = url_for(endpoint)

        entry = {
            "label": d["label"],
            "url": url,
            "active": path.startswith(url),
            "tooltip": d.get("tooltip", ""),
        }
        menu.append(entry)

    return menu



from werkzeug.utils import find_modules, import_string


def scan_modules(module_path, callback=None, recursive=True):
    for name in find_modules(module_path, include_packages=True, recursive=recursive):
        module = import_string(name)
        if callback:
            callback(module)



from pagic.scanner import scan_modules

MACROS = []


def macro(f):
    MACROS.append(f)
    return f


def register_macros(app, path):
    scan_modules(path)

    for macro in MACROS:
        app.template_global(macro.__name__)(macro)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pagic', 'tests']

package_data = \
{'': ['*'], 'tests': ['templates/tests/*']}

install_requires = \
['Flask>=2.0.2,<3.0.0']

setup_kwargs = {
    'name': 'pagic',
    'version': '0.2.1',
    'description': 'Top-level package for Pagic.',
    'long_description': '=====\nPagic\n=====\n\n\n.. image:: https://img.shields.io/pypi/v/pagic.svg\n        :target: https://pypi.python.org/pypi/pagic\n\n.. image:: https://img.shields.io/travis/abilian/pagic.svg\n        :target: https://travis-ci.com/abilian/pagic\n\n.. image:: https://readthedocs.org/projects/pagic/badge/?version=latest\n        :target: https://pagic.readthedocs.io/en/latest/?version=latest\n        :alt: Documentation Status\n\n\n\n\nPage-oriented magical web framework.\n\nNot ready for public consumption yet.\n\n\n* Free software: Apache Software License 2.0\n* Documentation: https://pagic.readthedocs.io.\n\n\nFeatures\n--------\n\n* TODO\n\nCredits\n-------\n\nThis package was created with Cruft_ and the `abilian/cookiecutter-abilian-python`_ project template.\n\n.. _Cruft: https://github.com/cruft/cruft\n.. _`abilian/cookiecutter-abilian-python`: https://github.com/abilian/cookiecutter-abilian-python\n',
    'author': 'Abilian SAS',
    'author_email': 'sf@abilian.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/abilian/pagic',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)
